@@ -9,7 +9,6 @@ import tensorflow as tf
 from datetime import datetime
 from pdb import set_trace
 from time import strftime, perf_counter, gmtime
-from prettytable import PrettyTable
 # My own module!
 import cvnn.layers as layers
 from cvnn.optimizers import get_optimizer
@@ -17,13 +16,30 @@ import cvnn.dataset as dp
 import cvnn.data_analysis as da
 from cvnn.utils import create_folder
 from cvnn import logger
+# Typing
+from cvnn.optimizers import t_optimizer
+from tensorflow.keras.losses import Loss
+from typing import Union, Optional, Tuple
+from tensorflow import Tensor, data
+from numpy import ndarray
+from cvnn.layers import t_layers_shape
 
+try:
+    from prettytable import PrettyTable
+    PRETTY_TABLE = True
+except ImportError:
+    PRETTY_TABLE = False
 
 try:
     import cPickle as pickle
 except ImportError:
     import pickle
 
+t_List = Union[ndarray, list]
+t_input_features = Union[t_List, Tensor, data.Dataset]
+t_labels = Union[t_List, Tensor]
+t_verbose = Union[str, int, bool]
+
 VERBOSITY = {0: "SILENT",  # verbosity 0. NADA DE NADA
              2: "FAST",    # verbosity 2. Muestra al final de cada epoch
              4: "PROBAR",  # Shows the progress bar but without acc or loss
@@ -49,8 +65,8 @@ class CvnnModel:
     # Constructor and Stuff
     # =====================
 
-    def __init__(self, name, shape, loss_fun, optimizer='Adam',
-                 verbose: bool = True, tensorboard: bool = True):
+    def __init__(self, name: str, shape: t_layers_shape, loss_fun: Loss, optimizer: t_optimizer = 'sgd',
+                 verbose: t_verbose = True, tensorboard: bool = True):
         """
         Constructor
         :param name: Name of the model. It will be used to distinguish models
@@ -107,7 +123,7 @@ class CvnnModel:
         self._manage_string(self.summary(), self._get_verbose(verbose), filename=self.name + "_metadata.txt", mode="x")
         self.plotter = da.Plotter(self.root_dir)
 
-    def call(self, x):
+    def call(self, x: t_input_features) -> Tensor:        # TODO: Beware, I think tf.Dataset will break here!
         """
         Forward result of the network
         :param x: Data input to be calculated
@@ -117,7 +133,7 @@ class CvnnModel:
             x = self.shape[i].call(x)
         return x
 
-    def _apply_loss(self, y_true, y_pred):
+    def _apply_loss(self, y_true: t_labels, y_pred: t_labels) -> Tensor:
         """
         Private! Use "evaluate loss" instead
         """
@@ -129,7 +145,7 @@ class CvnnModel:
                 sys.exit(-1)
         return tf.reduce_mean(input_tensor=self.loss_fun(y_true, y_pred), name=self.loss_fun.__name__)
 
-    def is_complex(self):
+    def is_complex(self) -> bool:
         """
         :return: True if the network is complex. False otherwise.
         """
@@ -260,7 +276,7 @@ class CvnnModel:
         return output_multiplier[1:]
 
     def get_real_equivalent(self, classifier: bool = True, capacity_equivalent: bool = True,
-                            equiv_technique: str = 'ratio', name: str = None):
+                            equiv_technique: str = 'ratio', name: Optional[str] = None):
         """
         Creates a new model equivalent of current model. If model is already real throws and error.
         :param classifier: True (default) if the model is a classification model. False otherwise.
@@ -319,7 +335,7 @@ class CvnnModel:
 
     # Add '@tf.function' to accelerate the code by a lot!
     @tf.function
-    def _train_step(self, x_train_batch, y_train_batch):
+    def _train_step(self, x_train_batch: Tensor, y_train_batch:  Tensor):
         """
         Performs one step of the training
         :param x_train_batch: input
@@ -350,10 +366,11 @@ class CvnnModel:
         with self.graph_writer.as_default():
             tf.summary.trace_export(name="graph", step=0, profiler_outdir=self.graph_writer_logdir)
 
-    def fit(self, x, y=None,
-            validation_split=0.0, validation_data=None, epochs: int = 10, batch_size: int = 32,
-            verbose=True, display_freq: int = 1,
-            save_model_checkpoints=False, save_csv_history=True, shuffle=True):
+    def fit(self, x: t_input_features, y: Optional[t_labels] = None,
+            validation_split: float = 0.0, validation_data: Optional[Union[Tuple[t_List], data.Dataset]] = None,
+            epochs: int = 10, batch_size: int = 32,
+            verbose: t_verbose = True, display_freq: int = 1,
+            save_model_checkpoints: bool = False, save_csv_history: bool = True, shuffle: bool = True):
         """
         Trains the model for a fixed number of epochs (iterations on a dataset).
 
@@ -398,7 +415,7 @@ class CvnnModel:
         :return: None
         """
         # Check input
-        verbose = self.verify_fit_input(save_model_checkpoints, epochs, batch_size, display_freq, verbose)
+        verbose = self._verify_fit_input(save_model_checkpoints, epochs, batch_size, display_freq, verbose)
         # Prepare dataset
         train_dataset, (x_test, y_test) = self._process_dataset(x, y, validation_split, validation_data, batch_size)
         # train_dataset = train_dataset.batch(batch_size=batch_size)  # TODO: Check if batch_size = 1
@@ -425,7 +442,7 @@ class CvnnModel:
             # Randomly shuffle the training data at the beginning of each epoch
             if shuffle:
                 train_dataset = train_dataset.shuffle(buffer_size=50000)
-            for x_batch, y_batch in train_dataset.prefetch(tf.data.experimental.AUTOTUNE).cache():
+            for x_batch, y_batch in train_dataset.prefetch(tf.data.experimental.AUTOTUNE):
                 if verbose in ("INFO", "DEBUG", "PROBAR"):
                     values = None
                     if verbose != "PROBAR":
@@ -475,7 +492,8 @@ class CvnnModel:
                             verbose)
         self.plotter.reload_data()
 
-    def verify_fit_input(self, save_model_checkpoints, epochs: int, batch_size: int, display_freq: int, verbose):
+    def _verify_fit_input(self, save_model_checkpoints: bool, epochs: int, batch_size: int, display_freq: int,
+                          verbose: t_verbose) -> str:
         assert not save_model_checkpoints  # TODO: Not working for the moment, sorry!
         if not (isinstance(epochs, int) and epochs > 0):
             logger.error("Epochs must be unsigned integer", exc_info=True)
@@ -509,7 +527,7 @@ class CvnnModel:
     """
 
     @staticmethod
-    def _get_verbose(verbose):
+    def _get_verbose(verbose: t_verbose) -> str:
         if isinstance(verbose, bool):
             return "INFO" if verbose else "SILENT"
         elif isinstance(verbose, int):
@@ -527,7 +545,31 @@ class CvnnModel:
             logger.error("verbose datatype (" + str(type(verbose)) + ") not supported")
 
     @staticmethod
-    def _process_dataset(x, y, validation_split=0.0, validation_data=None, batch_size: int = 32):
+    def _process_dataset(x: t_input_features, y: Optional[t_labels],
+                         validation_split: float = 0.0,
+                         validation_data: Optional[Union[Tuple[t_List], data.Dataset]] = None,
+                         batch_size: int = 32) -> Tuple[data.Dataset, Tuple]:
+        """
+        Process dataset and returns in the preferred format.
+
+        :param x: Input data. It could be:
+            - A Numpy array (or array-like), or a list of arrays (in case the model has multiple inputs).
+            - A TensorFlow tensor, or a list of tensors (in case the model has multiple inputs).
+            - A tf.data dataset. Should return a tuple (inputs, targets). Preferred data type (less overhead).
+        :param y: Labels/Target data. Like the input data x, it could be either Numpy array(s) or TensorFlow tensor(s).
+            If f x is a dataset then y will be ignored (default None)
+        :param validation_split: Float between 0 and 1.
+            Percentage of the input data to be used as test set (the rest will be use as train set)
+            Default: 0.0 (No validation set).
+            This input is ignored if validation_data is given.
+        :param validation_data: Data on which to evaluate the loss and any model metrics at the end of each epoch.
+            The model will not be trained on this data. This parameter takes precedence over validation_split.
+            It can be:
+                - tuple (x_val, y_val) of Numpy arrays or tensors. Preferred data type (less overhead).
+                - A tf.data dataset.
+        :param batch_size: (uint) Batch size of the data. Default 32 (because keras use 32 so... why not?)
+        :returns: (train, test) with train a tf.data.Dataset and test a tuple (x_val, y_val)
+        """
         test_dataset = None
         if isinstance(x, (list, tuple, np.ndarray)):
             if validation_data is None:
@@ -539,12 +581,12 @@ class CvnnModel:
                 y_test = y[:int(dataset_length * validation_split)]
                 if len(x_test) != 0:
                     validation_data = (np.array(x_test), np.array(y_test))
-                train_dataset = tf.data.Dataset.from_tensor_slices((x_train, y_train)).batch(batch_size=batch_size)
+                train_dataset = tf.data.Dataset.from_tensor_slices((x_train, y_train)).cache().batch(batch_size=batch_size)
             else:
                 if validation_split != 0:
                     logger.warning("validation_split was given but will be ignored because "
                                    "validation_data was not None")
-                train_dataset = tf.data.Dataset.from_tensor_slices((x, y)).batch(batch_size=batch_size)
+                train_dataset = tf.data.Dataset.from_tensor_slices((x, y)).cache().batch(batch_size=batch_size)
         elif isinstance(x, tf.data.Dataset):
             if y is not None:
                 logger.warning("y is ignored because x was a Dataset (and should contain the labels), "
@@ -571,12 +613,12 @@ class CvnnModel:
         assert isinstance(test_dataset, (list, tuple, np.ndarray)) or test_dataset is None
         if test_dataset is None:
             test_dataset = (None, None)
-        return train_dataset, test_dataset
+        return train_dataset.cache(), test_dataset
 
     @staticmethod
-    def _dataset_to_array(x, y=None):
+    def _dataset_to_array(x: t_input_features, y: Optional[t_labels] = None) -> Tuple[ndarray, ndarray]:
         if isinstance(y, (list, tuple, np.ndarray, tf.Tensor)):
-            return x, y
+            return np.array(x), np.array(y)
         elif isinstance(x, tf.data.Dataset):
             x_test = []
             y_test = []
@@ -596,7 +638,7 @@ class CvnnModel:
     # Predict models and results
     # ==========================
 
-    def predict(self, x):
+    def predict(self, x: t_input_features) -> Tensor:
         """
         Predicts the value of the class.
         ATTENTION: Use this only for classification tasks. For regression use "call" method.
@@ -606,7 +648,7 @@ class CvnnModel:
         y_out = self.call(x)
         return tf.math.argmax(y_out, 1)
 
-    def evaluate_loss(self, x, y=None):
+    def evaluate_loss(self, x: t_input_features, y: Optional[t_labels] = None) -> ndarray:
         """
         Computes the output of x and computes the loss using y
         :param x: Input of the netwotk
@@ -616,7 +658,7 @@ class CvnnModel:
         x, y = self._dataset_to_array(x, y)
         return self._apply_loss(y, self.call(x)).numpy()
 
-    def evaluate_accuracy(self, x, y=None):
+    def evaluate_accuracy(self, x: t_input_features, y: Optional[t_labels] = None) -> ndarray:
         """
         Computes the output of x and returns the accuracy using y as labels
         :param x: Input of the netwotk
@@ -659,6 +701,7 @@ class CvnnModel:
         filename = None
         if save_result:
             filename = self.root_dir / "categorical.csv"
+        # set_trace()
         return da.confusion_matrix(self.call(x), y, filename=filename)
 
     # ===========
@@ -753,7 +796,7 @@ class CvnnModel:
                                                   self.activation_summary_writer, self.epochs_done)
         self._save_tensorboard_gradients(x_train, y_train)
 
-    def _save_tensorboard_gradients(self, x_train, y_train):
+    def _save_tensorboard_gradients(self, x_train, y_train) -> None:
         with tf.GradientTape() as tape:
             x_called = self.call(x_train)  # Forward mode computation
             current_loss = self._apply_loss(y_train, x_called)  # Compute loss
@@ -783,7 +826,7 @@ class CvnnModel:
                     logger.error("Input_dtype not supported. Should never have gotten here!", exc_info=True)
                     sys.exit(-1)
 
-    def save(self, loss, acc):
+    def save(self, loss, acc) -> None:
         # https://stackoverflow.com/questions/2709800/how-to-pickle-yourself
         # TODO: TypeError: can't pickle _thread._local objects
         # https://github.com/tensorflow/tensorflow/issues/33283
@@ -805,7 +848,7 @@ class CvnnModel:
     # Managing strings
     # ================
 
-    def _manage_string(self, string, verbose="SILENT", filename=None, mode="a"):
+    def _manage_string(self, string, verbose="SILENT", filename=None, mode="a") -> None:
         """
         Prints a string to console and/or saves it to a file
         :param string: String to be printed/saved
@@ -833,8 +876,9 @@ class CvnnModel:
         return self._get_loss_and_acc_string(epoch, epochs, train_loss, train_acc, test_loss, test_acc, batch, batches)
 
     @staticmethod
-    def _get_loss_and_acc_string(epoch, epochs, train_loss, train_acc, test_loss=None, test_acc=None, batch=None,
-                                 batches=None) -> str:
+    def _get_loss_and_acc_string(epoch: int, epochs: int, train_loss: float, train_acc: float,
+                                 test_loss: float = None, test_acc: float = None, batch: int = None,
+                                 batches: int = None) -> str:
         ret_str = "Epoch {0}/{1}".format(epoch, epochs)
         if batch is not None and batches is not None:
             ret_str += "; batch {0}/{1}".format(batch, batches)
@@ -844,7 +888,7 @@ class CvnnModel:
             ret_str += "- test loss: {0:.4f} - test accuracy: {1:.2f} %".format(test_loss, test_acc * 100)
         return ret_str
 
-    def summary(self):
+    def summary(self) -> str:
         """
         Generates a string of a summary representation of your model.
         :return: string of the summary of the model
@@ -855,11 +899,21 @@ class CvnnModel:
             summary_str += "Complex Network\n"
         else:
             summary_str += "Real Network\n"
+        summary_str += "Loss function: " + self.loss_fun.__name__ + "\n"
+        summary_str += self.optimizer.summary()
         for lay in self.shape:
             summary_str += lay.get_description()
         return summary_str
 
-    def training_param_summary(self):
+    def training_param_summary(self) -> None:
+        """
+        Prints a table analog to tf.keras.Model.summary()
+        https://www.tensorflow.org/api_docs/python/tf/keras/Model#summary
+        """
+        if not PRETTY_TABLE:
+            logger.warning("Function CvnnModel.training_param_summary() was called "
+                           "but PrettyTable is not installed so it will be omitted")
+            return None
         summary_str = ""
         summary_str += self.name + "\n"
         if self.is_complex():
@@ -915,8 +969,9 @@ if __name__ == '__main__':
     # Train model
     model = CvnnModel("Testing_dropout", shape, tf.keras.losses.categorical_crossentropy,
                       tensorboard=False, verbose=False)
-    model.fit(x_fit, dataset.y, validation_split=0.0, batch_size=100, epochs=10,
+    model.fit(x_fit, dataset.y, validation_split=0.0, batch_size=100, epochs=3,
               verbose=True, save_csv_history=True)
+    model.call(x_fit)
     # start = time.time()
     # model.fit(dataset.x, dataset.y, batch_size=100, epochs=30, verbose=False)
     # end = time.time()
@@ -934,7 +989,7 @@ __author__ = 'J. Agustin BARRACHINA'
 __copyright__ = 'Copyright 2020, {project_name}'
 __credits__ = ['{credit_list}']
 __license__ = '{license}'
-__version__ = '0.2.49'
+__version__ = '0.2.57'
 __maintainer__ = 'J. Agustin BARRACHINA'
 __email__ = 'joseagustin.barra@gmail.com; jose-agustin.barrachina@centralesupelec.fr'
 __status__ = '{dev_status}'
